@@ -6,13 +6,42 @@
 -- SPDX-License-Identifier: MIT
 
 -- Receive script arguments from config.lua
-local config = ...
+local config = ... or {}
 
 -- ensure config.move and config.follow are not nil
 config.move = config.move or false
 config.follow = config.follow or false
+config.filter_forward_format = config["filter.forward-format"] or false
 
-local pending_rescan = false
+local self = {}
+self.scanning = false
+self.pending_rescan = false
+self.events_skipped = false
+self.pending_error_timer = nil
+
+function rescan()
+  for si in linkables_om:iterate() do
+    handleLinkable (si)
+  end
+end
+
+function scheduleRescan ()
+  if self.scanning then
+    self.pending_rescan = true
+    return
+  end
+
+  self.scanning = true
+  rescan ()
+  self.scanning = false
+
+  if self.pending_rescan then
+    self.pending_rescan = false
+    Core.sync(function ()
+      scheduleRescan ()
+    end)
+  end
+end
 
 function parseBool(var)
   return var and (var:lower() == "true" or var == "1")
@@ -82,12 +111,16 @@ function createLink (si, si_target, passthrough, exclusive)
   -- activate
   si_link:activate (Feature.SessionItem.ACTIVE, function (l, e)
     if e then
-      Log.warning (l, "failed to activate si-standard-link: " .. tostring(e))
-      si_flags[si_id].peer_id = nil
+      Log.info (l, "failed to activate si-standard-link: " .. tostring(e))
+      if si_flags[si_id] ~= nil then
+        si_flags[si_id].peer_id = nil
+      end
       l:remove ()
     else
-      si_flags[si_id].failed_peer_id = nil
-      si_flags[si_id].failed_count = 0
+      if si_flags[si_id] ~= nil then
+        si_flags[si_id].failed_peer_id = nil
+        si_flags[si_id].failed_count = 0
+      end
       Log.info (l, "activated si-standard-link")
     end
   end)
@@ -234,37 +267,170 @@ end
 -- that is currently being handled
 function findDefinedTarget (properties)
   local metadata = config.move and metadata_om:lookup()
-  local target_id = metadata
-      and metadata:find(properties["node.id"], "target.node")
-      or properties["node.target"]
   local target_direction = getTargetDirection(properties)
+  local target_key
+  local target_value
+  local node_defined = false
 
-  if target_id and tonumber(target_id) then
-    local si_target = linkables_om:lookup {
-      Constraint { "node.id", "=", target_id },
-    }
-    if si_target and canLink (properties, si_target) then
-      return si_target
-    end
+  if properties["target.object"] ~= nil then
+    target_value = properties["target.object"]
+    target_key = "object.serial"
+    node_defined = true
+  elseif properties["node.target"] ~= nil then
+    target_value = properties["node.target"]
+    target_key = "node.id"
+    node_defined = true
   end
 
-  if target_id then
-    for si_target in linkables_om:iterate() do
-      local target_props = si_target.properties
-      if (target_props["node.name"] == target_id or
-          target_props["object.path"] == target_id) and
-          target_props["item.node.direction"] == target_direction and
-          canLink (properties, si_target) then
-        return si_target
+  if metadata then
+    local id = metadata:find(properties["node.id"], "target.object")
+    if id ~= nil then
+      target_value = id
+      target_key = "object.serial"
+      node_defined = false
+    else
+      id = metadata:find(properties["node.id"], "target.node")
+      if id ~= nil then
+        target_value = id
+        target_key = "node.id"
+        node_defined = false
       end
     end
   end
-  return nil
+
+  if target_value == "-1" then
+    return nil, false, node_defined
+  end
+
+  if target_value and tonumber(target_value) then
+    local si_target = linkables_om:lookup {
+      Constraint { target_key, "=", target_value },
+    }
+    if si_target and canLink (properties, si_target) then
+      return si_target, true, node_defined
+    end
+  end
+
+  if target_value then
+    for si_target in linkables_om:iterate() do
+      local target_props = si_target.properties
+      if (target_props["node.name"] == target_value or
+          target_props["object.path"] == target_value) and
+          target_props["item.node.direction"] == target_direction and
+          canLink (properties, si_target) then
+        return si_target, true, node_defined
+      end
+    end
+  end
+  return nil, (target_value ~= nil), node_defined
 end
 
--- User or client defined target is not available, Loop through all the valid
--- linkables(nodes) and pick an appropriate one.
-function findUndefinedTarget (si)
+function parseParam(param, id)
+  local route = param:parse()
+  if route.pod_type == "Object" and route.object_id == id then
+    return route.properties
+  else
+    return nil
+  end
+end
+
+function arrayContains(a, value)
+  for _, v in ipairs(a) do
+    if v == value then
+      return true
+    end
+  end
+  return false
+end
+
+
+-- Does the target device have any active/available paths/routes to
+-- the physical device(spkr/mic/cam)?
+function haveAvailableRoutes (si_props)
+  local card_profile_device = si_props["card.profile.device"]
+  local device_id = si_props["device.id"]
+  local device = device_id and devices_om:lookup {
+    Constraint { "bound-id", "=", device_id, type = "gobject"},
+  }
+
+  if not card_profile_device or not device then
+    return true
+  end
+
+  local found = 0
+  local avail = 0
+
+  -- First check "SPA_PARAM_Route" if there are any active devices
+  -- in an active profile.
+  for p in device:iterate_params("Route") do
+    local route = parseParam(p, "Route")
+    if not route then
+      goto skip_route
+    end
+
+    if (route.device ~= tonumber(card_profile_device)) then
+      goto skip_route
+    end
+
+    if (route.available == "no") then
+      return false
+    end
+
+    do return true end
+
+    ::skip_route::
+  end
+
+  -- Second check "SPA_PARAM_EnumRoute" if there is any route that
+  -- is available if not active.
+  for p in device:iterate_params("EnumRoute") do
+    local route = parseParam(p, "EnumRoute")
+    if not route then
+      goto skip_enum_route
+    end
+
+    if not arrayContains(route.devices, tonumber(card_profile_device)) then
+      goto skip_enum_route
+    end
+    found = found + 1;
+    if (route.available ~= "no") then
+      avail = avail +1
+    end
+    ::skip_enum_route::
+  end
+
+  if found == 0 then
+    return true
+  end
+  if avail > 0 then
+    return true
+  end
+
+  return false
+
+end
+
+function findDefaultLinkable (si)
+  local si_props = si.properties
+  local target_direction = getTargetDirection(si_props)
+  local def_node_id = getDefaultNode(si_props, target_direction)
+  return linkables_om:lookup {
+      Constraint { "node.id", "=", tostring(def_node_id) }
+  }
+end
+
+function checkPassthroughCompatibility (si, si_target)
+  local si_must_passthrough = parseBool(si.properties["item.node.encoded-only"])
+  local si_target_must_passthrough = parseBool(si_target.properties["item.node.encoded-only"])
+  local can_passthrough = canPassthrough(si, si_target)
+  if (si_must_passthrough or si_target_must_passthrough)
+      and not can_passthrough then
+    return false, can_passthrough
+  end
+  return true, can_passthrough
+end
+
+function findBestLinkable (si)
   local si_props = si.properties
   local target_direction = getTargetDirection(si_props)
   local target_picked = nil
@@ -279,6 +445,7 @@ function findUndefinedTarget (si)
   } do
     local si_target_props = si_target.properties
     local si_target_node_id = si_target_props["node.id"]
+    local priority = tonumber(si_target_props["priority.session"]) or 0
 
     Log.debug(string.format("Looking at: %s (%s)",
         tostring(si_target_props["node.name"]),
@@ -289,28 +456,15 @@ function findUndefinedTarget (si)
       goto skip_linkable
     end
 
-    local priority = tonumber(si_target_props["priority.session"]) or 0
-
-    -- Is this linkable(node) a default one?
-    local def_node_id = getDefaultNode(si_props, target_direction)
-    if tostring(def_node_id) == si_target_node_id then
-      Log.debug(string.format("... this (%s) is the default node for %s",
-          def_node_id, target_direction))
-      priority = priority + 10000
+    if not haveAvailableRoutes(si_target_props) then
+      Log.debug("... does not have routes, skip linkable")
+      goto skip_linkable
     end
 
-    -- todo:check if this linkable(node/device) have valid routes.
-
-    -- Is passthrough feasible between these linkables(nodes)?
-    local si_must_passthrough = parseBool(si_props["item.node.encoded-only"])
-    local si_target_must_passthrough = parseBool(si_target_props["item.node.encoded-only"])
-    local can_passthrough = canPassthrough(si, si_target)
-
-    if (si_must_passthrough or si_target_must_passthrough)
-        and not can_passthrough then
-      Log.debug(string.format("... cannot passthrough, skip; must:%s target_must:%s",
-          tostring(si_must_passthrough),
-          tostring(si_target_must_passthrough)))
+    local passthrough_compatible, can_passthrough =
+        checkPassthroughCompatibility (si, si_target)
+    if not passthrough_compatible then
+      Log.debug("... passthrough is not compatible, skip linkable")
       goto skip_linkable
     end
 
@@ -336,7 +490,7 @@ function findUndefinedTarget (si)
   end
 
   if target_picked then
-    Log.info(string.format("... target: %s (%s), can_passthrough:%s",
+    Log.info(string.format("... best target picked: %s (%s), can_passthrough:%s",
       tostring(target_picked.properties["node.name"]),
       tostring(target_picked.properties["node.id"]),
       tostring(target_can_passthrough)))
@@ -347,6 +501,32 @@ function findUndefinedTarget (si)
 
 end
 
+function findUndefinedTarget (si)
+  -- Just find the best linkable if default nodes module is not loaded
+  if default_nodes == nil then
+    return findBestLinkable (si)
+  end
+
+  -- Otherwise find the default linkable. If the default linkable is not
+  -- compatible, we find the best one instead. We return nil if the default
+  -- linkable does not exist.
+  local si_target = findDefaultLinkable (si)
+  if si_target then
+    local passthrough_compatible, can_passthrough =
+        checkPassthroughCompatibility (si, si_target)
+    if canLink (si.properties, si_target) and passthrough_compatible then
+      Log.info(string.format("... default target picked: %s (%s), can_passthrough:%s",
+        tostring(si_target.properties["node.name"]),
+        tostring(si_target.properties["node.id"]),
+        tostring(can_passthrough)))
+      return si_target, can_passthrough
+    else
+      return findBestLinkable (si)
+    end
+  end
+  return nil, nil
+end
+
 function lookupLink (si_id, si_target_id)
   local link = links_om:lookup {
     Constraint { "out.item.id", "=", si_id },
@@ -361,16 +541,17 @@ function lookupLink (si_id, si_target_id)
   return link
 end
 
-function checkLinkable(si)
+function checkLinkable(si, handle_nonstreams)
   -- only handle stream session items
   local si_props = si.properties
-  if not si_props or si_props["item.node.type"] ~= "stream" then
+  if not si_props or (si_props["item.node.type"] ~= "stream"
+        and not handle_nonstreams)  then
     return false
   end
 
   -- Determine if we can handle item by this policy
-  local media_role = si_props["media.role"]
-  if endpoints_om:get_n_objects () > 0 and media_role ~= nil then
+  if endpoints_om:get_n_objects () > 0 and
+      si_props["item.factory.name"] == "si-audio-adapter" then
     return false
   end
 
@@ -379,7 +560,80 @@ end
 
 si_flags = {}
 
+function checkPending ()
+  local pending_linkables = pending_linkables_om:get_n_objects ()
+
+  -- We cannot process linkables if some of them are pending activation,
+  -- because linkables do not appear in the same order as nodes,
+  -- and we cannot resolve target node references until all linkables
+  -- have appeared.
+
+  if self.pending_error_timer then
+    self.pending_error_timer:destroy ()
+    self.pending_error_timer = nil
+  end
+
+  if pending_linkables ~= 0 then
+    -- Wait for linkables to get it sync
+    Log.debug(string.format("pending %d linkable not ready",
+        pending_linkables))
+    self.events_skipped = true
+
+    -- To make bugs in activation easier to debug, emit an error message
+    -- if they occur. policy-node should never be suspended for 20sec.
+    self.pending_error_timer = Core.timeout_add(20000, function()
+        self.pending_error_timer = nil
+        if pending_linkables ~= 0 then
+          Log.message(string.format("%d pending linkable(s) not activated in 20sec. "
+              .. "This should never happen.", pending_linkables))
+        end
+    end)
+
+    return true
+  elseif self.events_skipped then
+    Log.debug("pending linkables ready")
+    self.events_skipped = false
+    scheduleRescan ()
+    return true
+  end
+
+  return false
+end
+
+function checkFollowDefault (si, si_target, has_node_defined_target)
+  -- If it got linked to the default target that is defined by node
+  -- props but not metadata, start ignoring the node prop from now on.
+  -- This is what Pulseaudio does.
+  --
+  -- Pulseaudio skips here filter streams (i->origin_sink and
+  -- o->destination_source set in PA). Pipewire does not have a flag
+  -- explicitly for this, but we can use presence of node.link-group.
+  if not has_node_defined_target then
+    return
+  end
+
+  local si_props = si.properties
+  local target_props = si_target.properties
+  local reconnect = not parseBool(si_props["node.dont-reconnect"])
+  local is_filter = (si_props["node.link-group"] ~= nil)
+
+  if config.follow and default_nodes ~= nil and reconnect and not is_filter then
+    local def_id = getDefaultNode(si_props, getTargetDirection(si_props))
+
+    if target_props["node.id"] == tostring(def_id) then
+      local metadata = metadata_om:lookup()
+      -- Set target.node, for backward compatibility
+      metadata:set(tonumber(si_props["node.id"]), "target.node", "Spa:Id", "-1")
+      Log.info (si, "... set metadata to follow default")
+    end
+  end
+end
+
 function handleLinkable (si)
+  if checkPending () then
+    return
+  end
+
   local valid, si_props = checkLinkable(si)
   if not valid then
     return
@@ -395,10 +649,7 @@ function handleLinkable (si)
   Log.info (si, string.format("handling item: %s (%s)",
       tostring(si_props["node.name"]), tostring(si_props["node.id"])))
 
-  -- prepare flags table
-  if not si_flags[si.id] then
-    si_flags[si.id] = {}
-  end
+  ensureSiFlags(si)
 
   -- get other important node properties
   local reconnect = not parseBool(si_props["node.dont-reconnect"])
@@ -406,37 +657,29 @@ function handleLinkable (si)
   local si_must_passthrough = parseBool(si_props["item.node.encoded-only"])
 
   -- find defined target
-  local si_target = findDefinedTarget(si_props)
+  local si_target, has_defined_target, has_node_defined_target
+      = findDefinedTarget(si_props)
   local can_passthrough = si_target and canPassthrough(si, si_target)
 
   if si_target and si_must_passthrough and not can_passthrough then
     si_target = nil
   end
 
-  -- wait up to 2 seconds for the requested target to become available
-  -- this is because the client may have already "seen" a target that we haven't
-  -- yet prepared, which leads to a race condition
+  -- if the client has seen a target that we haven't yet prepared, schedule
+  -- a rescan one more time and hope for the best
   local si_id = si.id
-  if si_props["node.target"] and si_props["node.target"] ~= "-1"
+  if has_defined_target
       and not si_target
       and not si_flags[si_id].was_handled
       and not si_flags[si_id].done_waiting then
-    if not si_flags[si_id].timeout_source then
-      si_flags[si_id].timeout_source = Core.timeout_add(2000, function()
-        if si_flags[si_id] then
-          si_flags[si_id].done_waiting = true
-          si_flags[si_id].timeout_source = nil
-          rescan()
-        end
-        return false
-      end)
-    end
     Log.info (si, "... waiting for target")
+    si_flags[si_id].done_waiting = true
+    scheduleRescan()
     return
   end
 
   -- find fallback target
-  if not si_target then
+  if not si_target and (reconnect or not has_defined_target) then
     si_target, can_passthrough = findUndefinedTarget(si)
   end
 
@@ -444,21 +687,27 @@ function handleLinkable (si)
   if si_flags[si_id].peer_id then
     if si_target and si_flags[si_id].peer_id == si_target.id then
       Log.debug (si, "... already linked to proper target")
+      -- Check this also here, in case in default targets changed
+      checkFollowDefault (si, si_target, has_node_defined_target)
       return
     end
+    local link = lookupLink (si_id, si_flags[si_id].peer_id)
     if reconnect then
-      local link = lookupLink (si_id, si_flags[si_id].peer_id)
       if link ~= nil then
-        -- remove old link if active, otherwise schedule rescan
-        if ((link:get_active_features() & Feature.SessionItem.ACTIVE) ~= 0) then
-          si_flags[si_id].peer_id = nil
-          link:remove ()
-          Log.info (si, "... moving to new target")
-        else
-          pending_rescan = true
-          Log.info (si, "... scheduled rescan")
-          return
+        -- remove old link
+        if ((link:get_active_features() & Feature.SessionItem.ACTIVE) == 0) then
+          -- remove also not yet activated links: they might never become active,
+          -- and we should not loop waiting for them
+          Log.warning (link, "Link was not activated before removing")
         end
+        si_flags[si_id].peer_id = nil
+        link:remove ()
+        Log.info (si, "... moving to new target")
+      end
+    else
+      if link ~= nil then
+        Log.info (si, "... dont-reconnect, not moving")
+        return
       end
     end
   end
@@ -495,6 +744,9 @@ function handleLinkable (si)
     if not reconnect then
       Log.info (si, "... destroy node")
       node:request_destroy()
+    elseif si_flags[si.id].was_handled then
+      Log.info (si, "... waiting reconnect")
+      return
     end
 
     local client_id = node.properties["client.id"]
@@ -509,11 +761,13 @@ function handleLinkable (si)
   else
     createLink (si, si_target, can_passthrough, exclusive)
     si_flags[si.id].was_handled = true
+
+    checkFollowDefault (si, si_target, has_node_defined_target)
   end
 end
 
 function unhandleLinkable (si)
-  local valid, si_props = checkLinkable(si)
+  local valid, si_props = checkLinkable(si, true)
   if not valid then
     return
   end
@@ -530,7 +784,7 @@ function unhandleLinkable (si)
           si_flags[in_id] and si_flags[in_id].peer_id == out_id then
         si_flags[in_id].peer_id = nil
       elseif in_id == si.id and
-          si_flags[out_id] and si_flags[in_id].peer_id == in_id then
+          si_flags[out_id] and si_flags[out_id].peer_id == in_id then
         si_flags[out_id].peer_id = nil
       end
       silink:remove ()
@@ -541,20 +795,6 @@ function unhandleLinkable (si)
   si_flags[si.id] = nil
 end
 
-function rescan()
-  for si in linkables_om:iterate() do
-    handleLinkable (si)
-  end
-
-  -- if pending_rescan, re-evaluate after sync
-  if pending_rescan then
-    pending_rescan = false
-    Core.sync (function (c)
-      rescan()
-    end)
-  end
-end
-
 default_nodes = Plugin.find("default-nodes-api")
 
 metadata_om = ObjectManager {
@@ -568,11 +808,23 @@ endpoints_om = ObjectManager { Interest { type = "SiEndpoint" } }
 
 clients_om = ObjectManager { Interest { type = "client" } }
 
+devices_om = ObjectManager { Interest { type = "device" } }
+
 linkables_om = ObjectManager {
   Interest {
     type = "SiLinkable",
     -- only handle si-audio-adapter and si-node
     Constraint { "item.factory.name", "c", "si-audio-adapter", "si-node" },
+    Constraint { "active-features", "!", 0, type = "gobject" },
+  }
+}
+
+pending_linkables_om = ObjectManager {
+  Interest {
+    type = "SiLinkable",
+    -- only handle si-audio-adapter and si-node
+    Constraint { "item.factory.name", "c", "si-audio-adapter", "si-node" },
+    Constraint { "active-features", "=", 0, type = "gobject" },
   }
 }
 
@@ -584,39 +836,10 @@ links_om = ObjectManager {
   }
 }
 
-function cleanupTargetNodeMetadata()
-  local metadata = metadata_om:lookup()
-  if metadata then
-    local to_remove = {}
-    for s, k, t, v in metadata:iterate(Id.ANY) do
-      if k == "target.node" then
-        if v == "-1" then
-          -- target.node == -1 is useless, it means the default node
-          table.insert(to_remove, s)
-        else
-          -- if the target.node value is the same as the default node
-          -- that would be selected for this stream, remove it
-          local si = linkables_om:lookup { Constraint { "node.id", "=", s } }
-          local properties = si.properties
-          local def_id = getDefaultNode(properties, getTargetDirection(properties))
-          if tostring(def_id) == v then
-            table.insert(to_remove, s)
-          end
-        end
-      end
-    end
-
-    for _, s in ipairs(to_remove) do
-      metadata:set(s, "target.node", nil, nil)
-    end
-  end
-end
-
 -- listen for default node changes if config.follow is enabled
-if config.follow then
+if config.follow and default_nodes ~= nil then
   default_nodes:connect("changed", function ()
-    cleanupTargetNodeMetadata()
-    rescan()
+    scheduleRescan ()
   end)
 end
 
@@ -624,23 +847,135 @@ end
 if config.move then
   metadata_om:connect("object-added", function (om, metadata)
     metadata:connect("changed", function (m, subject, key, t, value)
-      if key == "target.node" then
-        rescan()
+      if key == "target.node" or key == "target.object" then
+        scheduleRescan ()
       end
     end)
   end)
 end
 
-linkables_om:connect("objects-changed", function (om)
-  rescan()
+function findAssociatedLinkGroupNode (si)
+  local si_props = si.properties
+  local node = si:get_associated_proxy ("node")
+  local link_group = node.properties["node.link-group"]
+  if link_group == nil then
+    return nil
+  end
+
+  -- get the associated media class
+  local assoc_direction = getTargetDirection(si_props)
+  local assoc_media_class =
+        si_props["media.type"] ..
+        (assoc_direction == "input" and "/Sink" or "/Source")
+
+  -- find the linkable with same link group and matching assoc media class
+  for assoc_si in linkables_om:iterate() do
+    local assoc_node = assoc_si:get_associated_proxy ("node")
+    local assoc_link_group = assoc_node.properties["node.link-group"]
+    if assoc_link_group == link_group and
+        assoc_media_class == assoc_node.properties["media.class"] then
+      return assoc_si
+    end
+  end
+
+  return nil
+end
+
+function onLinkGroupPortsStateChanged (si, old_state, new_state)
+  local new_str = tostring(new_state)
+  local si_props = si.properties
+
+  -- only handle items with configured ports state
+  if new_str ~= "configured" then
+    return
+  end
+
+  Log.info (si, "ports format changed on " .. si_props["node.name"])
+
+  -- find associated device
+  local si_device = findAssociatedLinkGroupNode (si)
+  if si_device ~= nil then
+    local device_node_name = si_device.properties["node.name"]
+
+    -- get the stream format
+    local f, m = si:get_ports_format()
+
+    -- unregister the device
+    Log.info (si_device, "unregistering " .. device_node_name)
+    si_device:remove()
+
+    -- set new format in the device
+    Log.info (si_device, "setting new format in " .. device_node_name)
+    si_device:set_ports_format(f, m, function (item, e)
+      if e ~= nil then
+        Log.warning (item, "failed to configure ports in " ..
+            device_node_name .. ": " .. e)
+      end
+
+      -- register back the device
+      Log.info (item, "registering " .. device_node_name)
+      item:register()
+    end)
+  end
+end
+
+function ensureSiFlags (si)
+  -- prepare flags table
+  if not si_flags[si.id] then
+    si_flags[si.id] = {}
+  end
+end
+
+function checkFiltersPortsState (si)
+  local si_props = si.properties
+  local node = si:get_associated_proxy ("node")
+  local link_group = node.properties["node.link-group"]
+
+  ensureSiFlags(si)
+
+  -- only listen for ports state changed on audio filter streams
+  if si_flags[si.id].ports_state_signal ~= true and
+      si_props["item.factory.name"] == "si-audio-adapter" and
+      si_props["item.node.type"] == "stream" and
+      link_group ~= nil then
+    si:connect("adapter-ports-state-changed", onLinkGroupPortsStateChanged)
+    si_flags[si.id].ports_state_signal = true
+    Log.info (si, "listening ports state changed on " .. si_props["node.name"])
+  end
+end
+
+linkables_om:connect("object-added", function (om, si)
+  local si_props = si.properties
+
+  -- Forward filters ports format to associated virtual devices if enabled
+  if config.filter_forward_format then
+    checkFiltersPortsState (si)
+  end
+
+  if si_props["item.node.type"] ~= "stream" then
+    scheduleRescan ()
+  else
+    handleLinkable (si)
+  end
 end)
 
 linkables_om:connect("object-removed", function (om, si)
   unhandleLinkable (si)
+  if si.properties["item.node.type"] ~= "stream" then
+    scheduleRescan ()
+  end
+end)
+
+devices_om:connect("object-added", function (om, device)
+  device:connect("params-changed", function (d, param_name)
+    scheduleRescan ()
+  end)
 end)
 
 metadata_om:activate()
 endpoints_om:activate()
 clients_om:activate()
 linkables_om:activate()
+pending_linkables_om:activate()
 links_om:activate()
+devices_om:activate()
